#
# Copyright (C) 2023 Python Packaging Authority
# Copyright (C) 2024 Sebastiano Barezzi
#
# SPDX-License-Identifier: MIT
#

import errno
import os

from sebaubuntu_libs.liblogging import LOGI

from .errors import DistutilsFileError, DistutilsInternalError

# cache for by mkpath() -- in addition to cheapening redundant calls,
# eliminates redundant "creating /foo/bar/baz" messages in dry-run mode
_path_created = {}


def mkpath(name, mode=0o777, verbose=1, dry_run=0):  # noqa: C901
	"""Create a directory and any missing ancestor directories.

	If the directory already exists (or if 'name' is the empty string, which
	means the current directory, which of course exists), then do nothing.
	Raise DistutilsFileError if unable to create some directory along the way
	(eg. some sub-path exists, but is a file rather than a directory).
	If 'verbose' is true, print a one-line summary of each mkdir to stdout.
	Return the list of directories actually created.

	os.makedirs is not used because:

	a) It's new to Python 1.5.2, and
	b) it blows up if the directory already exists (in which case it should
	   silently succeed).
	"""

	global _path_created

	# Detect a common bug -- name is None
	if not isinstance(name, str):
		raise DistutilsInternalError(f"mkpath: 'name' must be a string (got {name!r})")

	# XXX what's the better way to handle verbosity? print as we create
	# each directory in the path (the current behaviour), or only announce
	# the creation of the whole path? (quite easy to do the latter since
	# we're not using a recursive algorithm)

	name = os.path.normpath(name)
	created_dirs = []
	if os.path.isdir(name) or name == '':
		return created_dirs
	if _path_created.get(os.path.abspath(name)):
		return created_dirs

	(head, tail) = os.path.split(name)
	tails = [tail]  # stack of lone dirs to create

	while head and tail and not os.path.isdir(head):
		(head, tail) = os.path.split(head)
		tails.insert(0, tail)  # push next higher dir onto stack

	# now 'head' contains the deepest directory that already exists
	# (that is, the child of 'head' in 'name' is the highest directory
	# that does *not* exist)
	for d in tails:
		# print "head = %s, d = %s: " % (head, d),
		head = os.path.join(head, d)
		abs_head = os.path.abspath(head)

		if _path_created.get(abs_head):
			continue

		if verbose >= 1:
			LOGI("creating %s", head)

		if not dry_run:
			try:
				os.mkdir(head, mode)
			except OSError as exc:
				if not (exc.errno == errno.EEXIST and os.path.isdir(head)):
					raise DistutilsFileError(
						f"could not create '{head}': {exc.args[-1]}"
					)
			created_dirs.append(head)

		_path_created[abs_head] = 1
	return created_dirs

def copy_tree(  # noqa: C901
	src,
	dst,
	preserve_mode=1,
	preserve_times=1,
	preserve_symlinks=0,
	update=0,
	verbose=1,
	dry_run=0,
):
	"""Copy an entire directory tree 'src' to a new location 'dst'.

	Both 'src' and 'dst' must be directory names.  If 'src' is not a
	directory, raise DistutilsFileError.  If 'dst' does not exist, it is
	created with 'mkpath()'.  The end result of the copy is that every
	file in 'src' is copied to 'dst', and directories under 'src' are
	recursively copied to 'dst'.  Return the list of files that were
	copied or might have been copied, using their output name.  The
	return value is unaffected by 'update' or 'dry_run': it is simply
	the list of all files under 'src', with the names changed to be
	under 'dst'.

	'preserve_mode' and 'preserve_times' are the same as for
	'copy_file'; note that they only apply to regular files, not to
	directories.  If 'preserve_symlinks' is true, symlinks will be
	copied as symlinks (on platforms that support them!); otherwise
	(the default), the destination of the symlink will be copied.
	'update' and 'verbose' are the same as for 'copy_file'.
	"""
	from .file_util import copy_file

	if not dry_run and not os.path.isdir(src):
		raise DistutilsFileError("cannot copy tree '%s': not a directory" % src)
	try:
		names = os.listdir(src)
	except OSError as e:
		if dry_run:
			names = []
		else:
			raise DistutilsFileError(f"error listing files in '{src}': {e.strerror}")

	if not dry_run:
		mkpath(dst, verbose=verbose)

	outputs = []

	for n in names:
		src_name = os.path.join(src, n)
		dst_name = os.path.join(dst, n)

		if n.startswith('.nfs'):
			# skip NFS rename files
			continue

		if preserve_symlinks and os.path.islink(src_name):
			link_dest = os.readlink(src_name)
			if verbose >= 1:
				LOGI("linking %s -> %s", dst_name, link_dest)
			if not dry_run:
				os.symlink(link_dest, dst_name)
			outputs.append(dst_name)

		elif os.path.isdir(src_name):
			outputs.extend(
				copy_tree(
					src_name,
					dst_name,
					preserve_mode,
					preserve_times,
					preserve_symlinks,
					update,
					verbose=verbose,
					dry_run=dry_run,
				)
			)
		else:
			copy_file(
				src_name,
				dst_name,
				preserve_mode,
				preserve_times,
				update,
				verbose=verbose,
				dry_run=dry_run,
			)
			outputs.append(dst_name)

	return outputs



#
# Copyright (C) 2022 Sebastiano Barezzi
#
# SPDX-License-Identifier: Apache-2.0
#
"""sed (stream editor) library."""

import re

def sed(string: str, regexp: str, replacement: str, flags: str = ""):
	"""re wrapper for sed."""
	_flags = 0
	_flags |= (re.M if 'm' in flags or 'M' in flags else 0)
	_flags |= (re.I if 'i' in flags or 'I' in flags else 0)

	return re.sub(regexp, replacement, string, (0 if 'g' in flags else 1), _flags)



#
# Copyright (C) 2022 Sebastiano Barezzi
#
# SPDX-License-Identifier: Apache-2.0
#

from functools import cmp_to_key
from locale import strcoll

class Hal:
	"""Class representing a HAL."""
	def __init__(self, name) -> None:
		"""Initialize an object."""
		self.name = name

def strcoll_cast_to_str(obj1: object, obj2: object) -> int:
	obj1 = str(obj1)
	obj2 = str(obj2)

	return strcoll(obj1, obj2)

cast_to_str_key = cmp_to_key(strcoll_cast_to_str)



#
# Copyright (C) 2023 Python Packaging Authority
# Copyright (C) 2024 Sebastiano Barezzi
#
# SPDX-License-Identifier: MIT
#

class DistutilsError(Exception):
	"""The root of all Distutils evil."""

	pass

class DistutilsFileError(DistutilsError):
	"""Any problems in the filesystem: expected file not found, etc.
	Typically this is for problems that we detect before OSError
	could be raised."""

	pass

class DistutilsInternalError(DistutilsError):
	"""Internal inconsistencies or impossibilities (obviously, this
	should never be seen if the code is working!)."""

	pass



#
# Copyright (C) 2022 Sebastiano Barezzi
#
# SPDX-License-Identifier: Apache-2.0
#

from datetime import datetime
from io import BytesIO
from pathlib import Path
from sebaubuntu_libs.libgofile import raw_api
from sebaubuntu_libs.libgofile.account import Account
from sebaubuntu_libs.libgofile.contents import ContentResponse, Folder
from sebaubuntu_libs.libtyping import is_iterable_and_not_str
from typing import Iterable, Union

def get_server() -> str:
	"""Returns the best server available to receive files."""
	return raw_api.get_server()["server"]

def upload_file(file: Union[str, Path, BytesIO], server: str = None, token: str = None,
				folder_id: str = None, description: str = None, password: str = None,
				tags: Union[str, Iterable[str]] = None, expire: datetime = None):
	"""Upload one file on a specific server."""
	if server is None:
		server = get_server()

	needs_open = not isinstance(file, BytesIO)

	if is_iterable_and_not_str(tags):
		tags = ",".join(tags)

	if needs_open:
		with open(file, "rb") as f:
			data = raw_api.upload_file(f, server, token, folder_id, description, password, tags, expire)
	else:
		data = raw_api.upload_file(file, server, token, folder_id, description, password, tags, expire)

	return data

def get_content(content_id: str, token: str):
	"""Get a specific content details."""
	data = raw_api.get_content(content_id, token)

	return ContentResponse.from_dict(data)

def create_folder(parent_folder_id: str, folder_name: str, token: str):
	"""Create a new folder."""
	data = raw_api.create_folder(parent_folder_id, folder_name, token)

	return Folder.from_dict(data)

def set_folder_option(token: str, folder_id: str, option: str, value: str):
	"""Set an option on a folder."""
	raw_api.set_folder_option(token, folder_id, option, value)

	return True

def copy_content(contents_id: Union[str, Iterable[str]], folder_id_dest: str, token: str):
	"""Copy one or multiple contents to another folder."""
	if is_iterable_and_not_str(contents_id):
		contents_id = ",".join(contents_id)

	raw_api.copy_content(contents_id, folder_id_dest, token)

	return True

def delete_content(contents_id: Union[str, Iterable[str]], token: str):
	"""Delete one or multiple files/folders."""
	if is_iterable_and_not_str(contents_id):
		contents_id = ",".join(contents_id)

	raw_api.delete_content(contents_id, token)

	return True

def get_account(token: str):
	"""Get the account details."""
	# As of now toggling all_details does nothing
	data = raw_api.get_account_details(token, True)

	return Account.from_dict(data)



#
# Copyright (C) 2023 Python Packaging Authority
# Copyright (C) 2024 Sebastiano Barezzi
#
# SPDX-License-Identifier: MIT
#

import os.path

from .errors import DistutilsFileError

def _newer(source, target):
	return not os.path.exists(target) or (
		os.path.getmtime(source) > os.path.getmtime(target)
	)


def newer(source, target):
	"""
	Is source modified more recently than target.

	Returns True if 'source' is modified more recently than
	'target' or if 'target' does not exist.

	Raises DistutilsFileError if 'source' does not exist.
	"""
	if not os.path.exists(source):
		raise DistutilsFileError("file '%s' does not exist" % os.path.abspath(source))

	return _newer(source, target)



#
# Copyright (C) 2023 Python Packaging Authority
# Copyright (C) 2024 Sebastiano Barezzi
#
# SPDX-License-Identifier: MIT
#

def strtobool(val):
	"""Convert a string representation of truth to true (1) or false (0).

	True values are 'y', 'yes', 't', 'true', 'on', and '1'; false values
	are 'n', 'no', 'f', 'false', 'off', and '0'.  Raises ValueError if
	'val' is anything else.
	"""
	val = val.lower()
	if val in ('y', 'yes', 't', 'true', 'on', '1'):
		return 1
	elif val in ('n', 'no', 'f', 'false', 'off', '0'):
		return 0
	else:
		raise ValueError(f"invalid truth value {val!r}")



#
# Copyright (C) 2022 Sebastiano Barezzi
#
# SPDX-License-Identifier: Apache-2.0
#
"""AIK wrapper library."""

from git import Repo
from pathlib import Path
from platform import system
from sebaubuntu_libs.liblogging import LOGI
from shutil import which
from subprocess import check_output, STDOUT, CalledProcessError
from tempfile import TemporaryDirectory
from typing import Optional

AIK_REPO = "https://github.com/SebaUbuntu/AIK-Linux-mirror"

ALLOWED_OS = [
	"Linux",
	"Darwin",
]

class AIKImageInfo:
	def __init__(
		self,
		base_address: Optional[str],
		board_name: Optional[str],
		cmdline: Optional[str],
		dt: Optional[Path],
		dtb: Optional[Path],
		dtb_offset: Optional[str],
		dtbo: Optional[Path],
		header_version: Optional[str],
		image_type: Optional[str],
		kernel: Optional[Path],
		kernel_offset: Optional[str],
		origsize: Optional[str],
		os_version: Optional[str],
		pagesize: Optional[str],
		ramdisk: Optional[Path],
		ramdisk_compression: Optional[str],
		ramdisk_offset: Optional[str],
		sigtype: Optional[str],
		tags_offset: Optional[str],
	):
		self.kernel = kernel
		self.dt = dt
		self.dtb = dtb
		self.dtbo = dtbo
		self.ramdisk = ramdisk
		self.base_address = base_address
		self.board_name = board_name
		self.cmdline = cmdline
		self.dtb_offset = dtb_offset
		self.header_version = header_version
		self.image_type = image_type
		self.kernel_offset = kernel_offset
		self.origsize = origsize
		self.os_version = os_version
		self.pagesize = pagesize
		self.ramdisk_compression = ramdisk_compression
		self.ramdisk_offset = ramdisk_offset
		self.sigtype = sigtype
		self.tags_offset = tags_offset

	def __str__(self):
		return (
			f"base address: {self.base_address}\n"
			f"board name: {self.board_name}\n"
			f"cmdline: {self.cmdline}\n"
			f"dtb offset: {self.dtb_offset}\n"
			f"header version: {self.header_version}\n"
			f"image type: {self.image_type}\n"
			f"kernel offset: {self.kernel_offset}\n"
			f"original size: {self.origsize}\n"
			f"os version: {self.os_version}\n"
			f"page size: {self.pagesize}\n"
			f"ramdisk compression: {self.ramdisk_compression}\n"
			f"ramdisk offset: {self.ramdisk_offset}\n"
			f"sigtype: {self.sigtype}\n"
			f"tags offset: {self.tags_offset}\n"
		)

class AIKManager:
	"""
	This class is responsible for dealing with AIK tasks
	such as cloning, updating, and extracting recovery images.
	"""

	UNPACKING_FAILED_STRING = "Unpacking failed, try without --nosudo."

	def __init__(self):
		"""Initialize AIKManager class."""
		if system() not in ALLOWED_OS:
			raise NotImplementedError(f"{system()} is not supported")

		# Check whether cpio package is installed
		if which("cpio") is None:
			raise RuntimeError("cpio package is not installed")

		self.tempdir = TemporaryDirectory()
		self.path = Path(self.tempdir.name)

		self.images_path = self.path / "split_img"
		self.ramdisk_path = self.path / "ramdisk"

		LOGI("Cloning AIK...")
		Repo.clone_from(AIK_REPO, self.path)

	def unpackimg(self, image: Path, ignore_ramdisk_errors: bool = False):
		"""Extract recovery image."""
		image_prefix = image.name

		try:
			process = self._execute_script("unpackimg.sh", image)
		except CalledProcessError as e:
			returncode = e.returncode
			output = e.output
		else:
			returncode = 0
			output = process

		if returncode != 0:
			if self.UNPACKING_FAILED_STRING in output and ignore_ramdisk_errors:
				# Delete ramdisk folder to avoid issues
				try:
					self.ramdisk_path.rmdir()
				except Exception:
					pass
			else:
				raise RuntimeError(f"AIK extraction failed, return code {returncode}")

		return self._get_current_extracted_info(image_prefix)

	def repackimg(self):
		return self._execute_script("repack.sh")

	def cleanup(self):
		return self._execute_script("cleanup.sh")

	def _get_current_extracted_info(self, prefix: str):
		return AIKImageInfo(
			base_address=self._read_recovery_file(prefix, "base"),
			board_name=self._read_recovery_file(prefix, "board"),
			cmdline=self._read_recovery_file(prefix, "cmdline") \
				or self._read_recovery_file(prefix, "vendor_cmdline"),
			dt=self._get_extracted_info(prefix, "dt", check_size=True),
			dtb=self._get_extracted_info(prefix, "dtb", check_size=True),
			dtb_offset=self._read_recovery_file(prefix, "dtb_offset"),
			dtbo=self._get_extracted_info(prefix, "dtbo", check_size=True) \
				or self._get_extracted_info(prefix, "recovery_dtbo", check_size=True),
			header_version=self._read_recovery_file(prefix, "header_version", default="0"),
			image_type=self._read_recovery_file(prefix, "imgtype"),
			kernel=self._get_extracted_info(prefix, "kernel", check_size=True),
			kernel_offset=self._read_recovery_file(prefix, "kernel_offset"),
			origsize=self._read_recovery_file(prefix, "origsize"),
			os_version=self._read_recovery_file(prefix, "os_version"),
			pagesize=self._read_recovery_file(prefix, "pagesize"),
			ramdisk=self.ramdisk_path if self.ramdisk_path.is_dir() else None,
			ramdisk_compression=self._read_recovery_file(prefix, "ramdiskcomp") \
				or self._read_recovery_file(prefix, "vendor_ramdiskcomp"),
			ramdisk_offset=self._read_recovery_file(prefix, "ramdisk_offset"),
			sigtype=self._read_recovery_file(prefix, "sigtype"),
			tags_offset=self._read_recovery_file(prefix, "tags_offset"),
		)

	def _read_recovery_file(
		self, prefix: str, fragment: str, default: Optional[str] = None
	) -> Optional[str]:
		file = self._get_extracted_info(prefix, fragment)
		if not file:
			return default

		return file.read_text().splitlines()[0].strip()

	def _get_extracted_info(
		self, prefix: str, fragment: str, check_size: bool = False
	) -> Optional[Path]:
		path = self.images_path / f"{prefix}-{fragment}"

		if not path.is_file():
			return None

		try:
			if check_size and path.stat().st_size == 0:
				return None
		except Exception:
			return None

		return path

	def _execute_script(self, script: str, *args):
		command = [self.path / script, "--nosudo", *args]
		return check_output(command, stderr=STDOUT, universal_newlines=True, encoding="utf-8")



#
# Copyright (C) 2022 Sebastiano Barezzi
#
# SPDX-License-Identifier: Apache-2.0
#
"""Android fstab library."""

from itertools import repeat
from pathlib import Path
from typing import List, Set

from sebaubuntu_libs.libandroid.partitions.partition_model import PartitionModel

FSTAB_HEADER = "#<src>                                                 <mnt_point>            <type>  <mnt_flags and options>                            <fs_mgr_flags>\n"

class FstabEntry:
	"""
	A class representing a fstab entry
	"""
	def __init__(
		self,
		src: str,
		mount_point: str,
		fs_type: str,
		mnt_flags: List[str],
		fs_flags: List[str],
	):
		self.src = src
		self.mount_point = mount_point
		self.fs_type = fs_type
		self.mnt_flags = mnt_flags
		self.fs_flags = fs_flags

	def is_logical(self):
		return "logical" in self.fs_flags

	def is_slotselect(self):
		return "slotselect" in self.fs_flags

	@classmethod
	def from_entry(cls, line: str):
		src, mount_point, fs_type, mnt_flags, fs_flags = line.split()

		return cls(src, mount_point, fs_type, mnt_flags.split(','), fs_flags.split(','))

class Fstab:
	def __init__(self, fstab: Path):
		self.fstab = fstab

		self.entries: List[FstabEntry] = []

		for line in self.fstab.read_text().splitlines():
			if not line:
				continue

			if line.startswith("#"):
				continue

			self.entries.append(FstabEntry.from_entry(line))

	def __str__(self):
		return self.format()

	def format(self, twrp: bool = False):
		entries = []

		src_len_max = 0
		mount_point_len_max = 0
		fs_type_len_max = 0
		mnt_flags_len_max = 0

		for entry in self.entries:
			mount_point_len = len(entry.mount_point)
			if mount_point_len > mount_point_len_max:
				mount_point_len_max = mount_point_len

			fs_type_len = len(entry.fs_type)
			if fs_type_len > fs_type_len_max:
				fs_type_len_max = fs_type_len

			src_len = len(entry.src)
			if src_len > src_len_max:
				src_len_max = src_len

			mnt_flags_len = len(entry.mnt_flags)
			if mnt_flags_len > mnt_flags_len_max:
				mnt_flags_len_max = mnt_flags_len

		src_len_max += 5
		mount_point_len_max += 5
		fs_type_len_max += 5
		mnt_flags_len_max += 5

		for entry in self.entries:
			src_space = ""
			mount_point_space = ""
			fs_type_space = ""
			mnt_flags_space = ""

			for _ in repeat(None, src_len_max - len(entry.src)):
				src_space += " "
			for _ in repeat(None, mount_point_len_max - len(entry.mount_point)):
				mount_point_space += " "
			for _ in repeat(None, fs_type_len_max - len(entry.fs_type)):
				fs_type_space += " "
			for _ in repeat(None, mnt_flags_len_max - len(entry.mnt_flags)):
				mnt_flags_space += " "

			if not twrp:
				entries.append(f"{entry.src}{src_space}{entry.mount_point}{mount_point_space}{entry.fs_type}{fs_type_space}{','.join(entry.mnt_flags)}{mnt_flags_space}{','.join(entry.fs_flags)}")
			else:
				flags = [f'display={Path(entry.mount_point).name}']
				if entry.is_logical():
					flags.append("logical")
				if entry.is_slotselect():
					flags.append("slotselect")
				entries.append(f"{entry.mount_point}{mount_point_space}{entry.fs_type}{fs_type_space}{entry.src}{src_space}flags={';'.join(flags)}")

		entries.append("")

		return "\n".join(entries)

	def get_partition_by_mount_point(self, mount_point: str):
		for entry in self.entries:
			if entry.mount_point == mount_point:
				return entry

		return None

	def get_logical_partitions(self):
		return [entry for entry in self.entries if entry.is_logical()]

	def get_logical_partitions_models(self) -> Set[PartitionModel]:
		models = set()

		for entry in self.entries:
			if entry.is_logical():
				partition_model = PartitionModel.from_mount_point(entry.mount_point)
				if partition_model is None:
					continue
				models.add(partition_model)

		return models

	def get_slotselect_partitions(self):
		return [entry for entry in self.entries if entry.is_slotselect()]

	def get_ab_partitions_models(self) -> Set[PartitionModel]:
		models = set()

		for entry in self.get_slotselect_partitions():
			partition_model = PartitionModel.from_mount_point(entry.mount_point)
			if partition_model is None:
				continue
			models.add(partition_model)

		return models



#
# Copyright (C) 2022 Sebastiano Barezzi
#
# SPDX-License-Identifier: Apache-2.0
#

from pathlib import Path
from typing import TYPE_CHECKING

from sebaubuntu_libs.libandroid.elf.elf import ELF

if TYPE_CHECKING:
	from sebaubuntu_libs.libandroid.partitions.partition import AndroidPartition
else:
	AndroidPartition = None

class SharedLibrary:
	def __init__(
		self,
		name: str,
		partition: AndroidPartition,
	):
		self.name = name
		self.partition = partition

		self.filename = f"{self.name}.so"

		path_32 = self.partition.path / "lib" / self.filename
		self.lib_32 = ELF(path_32) if path_32.is_file() else None

		path_64 = self.partition.path / "lib64" / self.filename
		self.lib_64 = ELF(path_64) if path_64.is_file() else None

		assert self.lib_32 or self.lib_64, f"Library {self.name} not found in {self.partition.path}"

		needed_libraries_32 = self.lib_32.needed_libraries if self.lib_32 else None
		needed_libraries_64 = self.lib_64.needed_libraries if self.lib_64 else None

		if needed_libraries_32 is not None and needed_libraries_64 is not None:
			if needed_libraries_32 != needed_libraries_64:
				raise ValueError(
					f"32-bit and 64-bit versions of {self.name} have different dependencies"
				)

		self.needed_libraries = (
			self.lib_64.needed_libraries
			if self.lib_64 is not None
			else self.lib_32.needed_libraries
			if self.lib_32 is not None
			else set()
		)

	def is_multilib(self) -> bool:
		"""Check if this library is multilib (both 32 and 64bit)."""
		return self.lib_32 is not None and self.lib_64 is not None

	def get_path(self) -> Path:
		"""Get either 64bit library path if exists, else 32bit one."""
		if self.lib_64:
			return self.lib_64.path
		elif self.lib_32:
			return self.lib_32.path
		else:
			raise ValueError("Library not found")



#
# Copyright (C) 2022 Sebastiano Barezzi
#
# SPDX-License-Identifier: Apache-2.0
#

from datetime import datetime
from typing import Dict, List

class Content:
	def __init__(self,
	             content_id: str,
				 content_type: str,
				 name: str,
				 parent_folder: str,
				 create_time: datetime,
				):
		"""Initialize a GoFile content."""
		self.content_id = content_id
		self.content_type = content_type
		self.name = name
		self.parent_folder = parent_folder
		self.create_time = create_time

	def get_kwargs(self):
		return {
			"content_id": self.content_id,
			"content_type": self.content_type,
			"name": self.name,
			"parent_folder": self.parent_folder,
			"create_time": self.create_time,
		}

	@staticmethod
	def from_dict(data: Dict):
		create_time = datetime.fromtimestamp(data["createTime"])

		return Content(content_id=data["id"], content_type=data["type"], name=data["name"],
		               parent_folder=data["parentFolder"], create_time=create_time)

class File(Content):
	"""Class representing a GoFile file."""
	def __init__(self,
	             size: int,
	             download_count: int,
				 md5: str,
				 mimetype: str,
				 server_choosen: str,
				 direct_link: str,
				 link: str,
				 *args, **kwargs,
	            ):
		"""Initialize a GoFile file."""
		super().__init__(*args, **kwargs)

		self.size = size
		self.download_count = download_count
		self.md5 = md5
		self.mimetype = mimetype
		self.server_choosen = server_choosen
		self.direct_link = direct_link
		self.link = link

	def get_kwargs(self):
		kwargs = super().get_kwargs()

		kwargs["size"] = self.size
		kwargs["download_count"] = self.download_count
		kwargs["md5"] = self.md5
		kwargs["mimetype"] = self.mimetype
		kwargs["server_choosen"] = self.server_choosen
		kwargs["direct_link"] = self.direct_link
		kwargs["link"] = self.link

		return kwargs

	@staticmethod
	def from_dict(data: Dict):
		content = Content.from_dict(data)

		return File(size=data["size"], download_count=data["downloadCount"], md5=data["md5"],
		            mimetype=data["mimetype"], server_choosen=data["serverChoosen"],
					direct_link=data["directLink"], link=data["link"],
					**content.get_kwargs())

class Folder(Content):
	"""Class representing a GoFile folder."""
	def __init__(self,
				 childs: List[str],
				 code: str,
				 public: bool,
				 *args, **kwargs,
				):
		"""Initialize a GoFile folder."""
		super().__init__(*args, **kwargs)

		self.childs = childs
		self.code = code
		self.public = public

	def get_kwargs(self):
		kwargs = super().get_kwargs()

		kwargs["childs"] = self.childs
		kwargs["code"] = self.code
		kwargs["public"] = self.public

		return kwargs

	@staticmethod
	def from_dict(data: Dict):
		content = Content.from_dict(data)

		public = data.get("public")

		return Folder(childs=data["childs"], code=data["code"], public=public,
					  **content.get_kwargs())

class ContentResponse(Folder):
	"""Class representing a GoFile content response.

	As of now all contents returned by get_content are folders."""
	def __init__(self,
				 total_download_count: int,
				 total_size: int,
				 contents: Dict[str, Dict],
				 # We may not know the owner ID if the owner isn't the one who called get_content
				 owner_id: str = None,
				 # Only for non-root folders
				 parent_folder: str = None,
				 # Only for root folders
				 is_root: bool = False,
				 *args, **kwargs,
				):
		"""Initialize a GoFile content."""
		super().__init__(*args, **kwargs)

		self.total_download_count = total_download_count
		self.total_size = total_size
		self.contents = contents
		self.owner_id = owner_id
		self.parent_folder = parent_folder
		self.is_root = is_root

	@staticmethod
	def from_dict(data: Dict):
		folder = Folder.from_dict(data)

		owner_id = data.get("ownerId")
		parent_folder = data.get("parentFolder")
		is_root = data.get("isRoot", False)

		return ContentResponse(total_download_count=data["totalDownloadCount"],
		                       total_size=data["totalSize"], contents=data["contents"],
							   owner_id=owner_id, parent_folder=parent_folder,
							   is_root=is_root,
							   **folder.get_kwargs())



#
# Copyright (C) 2022 Sebastiano Barezzi
#
# SPDX-License-Identifier: Apache-2.0
#

from elftools.common.exceptions import ELFError
from elftools.elf.elffile import ELFFile
from pathlib import Path
from typing import Set

class ELF:
	def __init__(self, path: Path):
		self.path = path

		# Just check that this is actually an ELF file, die otherwise
		with self.path.open("rb") as f:
			ELFFile(f)

		self.needed_libraries = self.get_needed_libs(self.path)

	@classmethod
	def get_needed_libs(cls, file: Path) -> Set[str]:
		needed_libs: Set[str] = set()

		with file.open("rb") as f:
			try:
				elf = ELFFile(f)
				dynsec = elf.get_section_by_name(".dynamic")
				if dynsec:
					for dt_needed in dynsec.iter_tags("DT_NEEDED"):
						needed_libs.add(str(dt_needed.needed))
			except ELFError:
				pass

		return needed_libs



#
# Copyright (C) 2022 Sebastiano Barezzi
#
# SPDX-License-Identifier: Apache-2.0
#

from enum import Enum
from typing import Any, Callable, List

from sebaubuntu_libs.libandroid.props import BuildProp
from sebaubuntu_libs.libandroid.props.utils import fingerprint_to_description, get_partition_props
from sebaubuntu_libs.libcompat.distutils.util import strtobool

def get_product_props(value: str):
	return get_partition_props("ro.product.{}" + value, add_empty=True)

DEVICE_CODENAME = get_product_props("device")
DEVICE_MANUFACTURER = get_product_props("manufacturer")
DEVICE_BRAND = get_product_props("brand")
DEVICE_MODEL = get_product_props("model")

DEVICE_ARCH = ["ro.bionic.arch"]
DEVICE_CPU_ABILIST = get_partition_props("ro.{}product.cpu.abilist", add_empty=True)
DEVICE_CPU_VARIANT = ["ro.bionic.cpu_variant"]
DEVICE_SECOND_ARCH = ["ro.bionic.2nd_arch"]
DEVICE_SECOND_CPU_VARIANT = ["ro.bionic.2nd_cpu_variant"]

DEVICE_IS_AB = ["ro.build.ab_update"]
DEVICE_USES_DYNAMIC_PARTITIONS = ["ro.boot.dynamic_partitions"]
DEVICE_USES_VIRTUAL_AB = ["ro.virtual_ab.enabled"]
DEVICE_USES_SYSTEM_AS_ROOT = ["ro.build.system_root_image"]

BOOTLOADER_BOARD_NAME = ["ro.product.board"]
DEVICE_PLATFORM = ["ro.board.platform"]
DEVICE_PIXEL_FORMAT = ["ro.minui.pixel_format"]
SCREEN_DENSITY = ["ro.sf.lcd_density"]
USE_VULKAN = ["ro.hwui.use_vulkan"]

BUILD_FINGERPRINT = get_partition_props("ro.{}build.fingerprint", add_empty=True)
BUILD_DESCRIPTION = get_partition_props("ro.{}build.description", add_empty=True)

GMS_CLIENTID_BASE = ["ro.com.google.clientidbase.ms", "ro.com.google.clientidbase"]

BUILD_SECURITY_PATCH = ["ro.build.version.security_patch"]
BUILD_VENDOR_SECURITY_PATCH = ["ro.vendor.build.security_patch"]

FIRST_API_LEVEL = ["ro.product.first_api_level"]
PRODUCT_CHARACTERISTICS = ["ro.build.characteristics"]
APEX_UPDATABLE = ["ro.apex.updatable"]

BOARD_FIRST_API_LEVEL = ["ro.board.first_api_level"]
BOARD_API_LEVEL = ["ro.board.api_level"]

class DeviceArch(Enum):
	def __new__(cls, *args, **kwargs):
		value = len(cls.__members__) + 1
		obj = object.__new__(cls)
		obj._value_ = value
		return obj

	def __init__(self,
	             arch: str,
	             arch_variant: str,
	             bitness: int,
	             cpu_abilist: List[str],
	            ):
		self.arch = arch
		self.arch_variant = arch_variant
		self.bitness = bitness
		self.cpu_abilist = cpu_abilist

		self.cpu_abi = self.cpu_abilist[0]
		self.cpu_abi2 = self.cpu_abilist[1] if len(self.cpu_abilist) > 1 else ""

	def __bool__(self):
		return self.arch != "unknown"

	def __str__(self):
		return self.arch

	@classmethod
	def from_arch(cls, arch: str):
		for arch_enum in cls:
			if arch_enum.arch != arch:
				continue

			return arch_enum

		raise ValueError(f"Unknown arch: {arch}")

	@classmethod
	def from_abi(cls, abi: str):
		for arch_enum in cls:
			if abi not in arch_enum.cpu_abilist:
				continue

			return arch_enum

		raise ValueError(f"Unknown ABI: {abi}")

	ARM = ("arm", "armv7-a-neon", 32, ["armeabi-v7a", "armeabi"])
	ARM64 = ("arm64", "armv8-a", 64, ["arm64-v8a"])
	X86 = ("x86", "generic", 32, ["x86"])
	X86_64 = ("x86_64", "generic", 64, ["x86_64"])

bool_cast = lambda x: bool(strtobool(x))

class DeviceInfo:
	"""
	This class is responsible for reading parse common build props by using BuildProp class.
	"""

	def __init__(self, build_prop: BuildProp):
		"""
		Parse common build props.
		"""
		self.build_prop = build_prop

		# Parse props
		self.codename = self.get_first_prop(DEVICE_CODENAME)
		self.manufacturer = self.get_first_prop(DEVICE_MANUFACTURER).split()[0].lower()
		self.brand = self.get_first_prop(DEVICE_BRAND, raise_exception=False)
		self.model = self.get_first_prop(DEVICE_MODEL, raise_exception=False)
		self.build_fingerprint = self.get_first_prop(BUILD_FINGERPRINT, raise_exception=False)
		self.build_description = self.get_first_prop(
			BUILD_DESCRIPTION, default=fingerprint_to_description(self.build_fingerprint) if self.build_fingerprint else None
		)

		# Parse arch
		self.arch = None
		self.second_arch = None

		arch_prop = self.get_first_prop(DEVICE_ARCH, raise_exception=False)
		second_arch_prop = self.get_first_prop(DEVICE_SECOND_ARCH, raise_exception=False)
		if arch_prop:
			self.arch = DeviceArch.from_arch(arch_prop)
			if second_arch_prop:
				self.second_arch = DeviceArch.from_arch(second_arch_prop)
		else:
			# Fallback to ABI list
			abi_list = self.get_first_prop(DEVICE_CPU_ABILIST).split(",")
			assert abi_list, "No ABI list prop found"
			archs = list(set([DeviceArch.from_abi(abi) for abi in abi_list]))
			assert 0 < len(archs) <= 2, "Invalid ABI list"
			# Higher bitness architectures has priority
			archs.sort(key=lambda x: x.bitness, reverse=True)
			self.arch = archs[0]
			if len(archs) > 1:
				self.second_arch = archs[1]

		self.cpu_variant = self.get_first_prop(DEVICE_CPU_VARIANT, default="generic")
		self.second_cpu_variant = self.get_first_prop(DEVICE_SECOND_CPU_VARIANT, default="generic")

		self.bootloader_board_name = self.get_first_prop(BOOTLOADER_BOARD_NAME)
		self.platform = self.get_first_prop(DEVICE_PLATFORM, default="default")
		self.device_is_ab = self.get_first_prop(DEVICE_IS_AB, data_type=bool_cast, default=False)
		self.device_uses_dynamic_partitions = self.get_first_prop(DEVICE_USES_DYNAMIC_PARTITIONS, data_type=bool_cast, default=False)
		self.device_uses_virtual_ab = self.get_first_prop(DEVICE_USES_VIRTUAL_AB, data_type=bool_cast, default=False)
		self.device_uses_system_as_root = self.get_first_prop(DEVICE_USES_SYSTEM_AS_ROOT, data_type=bool_cast, default=False)
		self.device_uses_updatable_apex = self.get_first_prop(APEX_UPDATABLE, data_type=bool_cast, default=False)

		self.device_pixel_format = self.get_first_prop(DEVICE_PIXEL_FORMAT, raise_exception=False)
		self.screen_density = self.get_first_prop(SCREEN_DENSITY, raise_exception=False)
		self.use_vulkan = self.get_first_prop(USE_VULKAN, data_type=bool_cast, default=False)
		self.gms_clientid_base = self.get_first_prop(GMS_CLIENTID_BASE, default=f"android-{self.manufacturer}")
		self.first_api_level = self.get_first_prop(FIRST_API_LEVEL)
		self.product_characteristics = self.get_first_prop(PRODUCT_CHARACTERISTICS, default="")

		self.build_security_patch = self.get_first_prop(BUILD_SECURITY_PATCH)
		self.vendor_build_security_patch = self.get_first_prop(BUILD_VENDOR_SECURITY_PATCH, default=self.build_security_patch)

		self.board_first_api_level = self.get_first_prop(BOARD_FIRST_API_LEVEL, raise_exception=False)
		self.board_api_level = self.get_first_prop(BOARD_API_LEVEL, raise_exception=False)

	def get_first_prop(self, props: List[str], data_type: Callable[[str], Any] = str,
	                   default: Any = None, raise_exception: bool = True):
		for prop in props:
			prop_value = self.build_prop._get_prop(prop, data_type)
			if prop_value is None:
				continue

			return prop_value

		if default is None and raise_exception:
			raise AssertionError(f'Property {props[0]} could not be found in build.prop')
		else:
			return default



#
# Copyright (C) 2022 Sebastiano Barezzi
#
# SPDX-License-Identifier: Apache-2.0
#
"""Android build prop library."""

from __future__ import annotations
from pathlib import Path
from typing import Any, Callable, List, Optional, Union

from sebaubuntu_libs.libcompat.distutils.util import strtobool

class BuildProp(dict):
	"""
	A class representing a build prop.
	This class basically mimics Android props system, with both getprop and setprop commands
	"""
	@classmethod
	def from_file(cls, file: Path):
		"""Create a BuildProp object from a file."""
		build_prop = cls()
		build_prop.import_props(file)
		return build_prop

	def __str__(self):
		return self.get_readable_list()

	def get_readable_list(self, excluded_props: List[str] = []):
		ordered_props = dict(sorted(self.items()))

		for excluded_prop in excluded_props:
			ordered_props.pop(excluded_prop, None)

		return "\n".join(f"{key}={value}" for key, value in ordered_props.items()) + "\n"

	def import_props(self, file: Union[Path, BuildProp]):
		if isinstance(file, BuildProp):
			text = str(file)
		else:
			text = file.read_text(encoding="utf-8")

		for prop in text.splitlines():
			if prop.startswith("#"):
				continue
			try:
				prop_name, prop_value = prop.split("=", 1)
			except ValueError:
				continue
			else:
				self.set_prop(prop_name, prop_value)

	def _get_prop(self, key: str, data_type: Callable[[str], Any] = str, default: Any = None):
		if key in self:
			try:
				return data_type(self[key])
			except ValueError:
				return default
		else:
			return default

	def get_prop(self, key: str, default: Optional[str] = None) -> Optional[str]:
		return self._get_prop(key, str, default)

	def get_prop_bool(self, key: str, default: bool = False) -> bool:
		return self._get_prop(key, lambda x: bool(strtobool(x)), default)

	def get_prop_int(self, key: str, default: int = 0) -> int:
		return self._get_prop(key, int, default)

	def get_prop_float(self, key: str, default: float = 0.0) -> float:
		return self._get_prop(key, float, default)

	def set_prop(self, key: str, value: str):
		assert isinstance(value, str), f"value must be a string, not {type(value)}"
		self[key] = value

	def write_to_file(self, path: Path, excluded_props: List[str] = []):
		path.write_text(self.get_readable_list(excluded_props), encoding="utf-8")



#
# Copyright (C) 2022 Sebastiano Barezzi
#
# SPDX-License-Identifier: Apache-2.0
#

from __future__ import annotations
from pathlib import Path
from typing import List, Optional

from sebaubuntu_libs.libandroid.fstab import Fstab, FstabEntry
from sebaubuntu_libs.libandroid.partitions.partition_model import PartitionModel
from sebaubuntu_libs.libandroid.props import BuildProp
from sebaubuntu_libs.libandroid.vintf.manifest import Manifest
from sebaubuntu_libs.libreorder import strcoll_files_key

BUILD_PROP_LOCATION = ["build.prop", "etc/build.prop"]
DEFAULT_PROP_LOCATION = ["default.prop", "etc/default.prop"]

MANIFEST_LOCATION = ["manifest.xml", "etc/vintf/manifest.xml"]

def get_files_list(path: Path) -> List[Path]:
	files = []

	for i in path.iterdir():
		if i.is_file():
			files.append(i)
		elif i.is_dir():
			files.extend(get_files_list(i))

	return files

class AndroidPartition:
	def __init__(self, model: PartitionModel, path: Path):
		self.model = model
		self.path = path

		self.files = get_files_list(self.path)

		self.fstab_entry: Optional[FstabEntry] = None

		self.build_prop = BuildProp()
		for possible_paths in BUILD_PROP_LOCATION + DEFAULT_PROP_LOCATION:
			build_prop_path = self.path / possible_paths
			if not build_prop_path.is_file():
				continue

			self.build_prop.import_props(build_prop_path)

		self.manifest = Manifest()
		for possible_paths in MANIFEST_LOCATION:
			manifest_path = self.path / possible_paths
			if not manifest_path.is_file():
				continue

			self.manifest.import_file(manifest_path)

	def fill_fstab_entry(self, fstab: Fstab):
		for mount_point in self.model.mount_points:
			self.fstab_entry = fstab.get_partition_by_mount_point(mount_point)
			if self.fstab_entry is not None:
				return

	def get_files(self):
		"""Returns the ordered list of files."""
		self.files.sort(key=strcoll_files_key)
		return self.files

	def get_formatted_file(self, file: Path):
		return self.model.proprietary_files_prefix / file.relative_to(self.path)

	def get_formatted_files(self):
		return [self.get_formatted_file(file) for file in self.get_files()]



#
# Copyright (C) 2022 Sebastiano Barezzi
#
# SPDX-License-Identifier: Apache-2.0
#

from enum import IntEnum
from pathlib import Path
from typing import List, Optional

class PartitionGroup(IntEnum):
	BOOTLOADER = 0
	SSI = 1
	TREBLE = 2
	DATA = 3

(
	BOOTLOADER,
	SSI,
	TREBLE,
	DATA,
) = PartitionGroup.__members__.values()

class _PartitionModel:
	ALL: List["_PartitionModel"] = []

	def __init__(
		self,
		name: str,
		group: PartitionGroup,
		mount_points: Optional[List[str]] = None,
		proprietary_files_prefix: Optional[Path] = None,
	):
		self.name = name
		self.group = group
		self.mount_points = mount_points or [f"/{self.name}"]
		self.proprietary_files_prefix = proprietary_files_prefix or Path(self.name)

		_PartitionModel.ALL.append(self)

	@classmethod
	def from_name(cls, name: str):
		for model in cls.ALL:
			if model.name == name:
				return model

		return None

	@classmethod
	def from_group(cls, group: PartitionGroup):
		return [model for model in cls.ALL if model.group == group]

	@classmethod
	def from_mount_point(cls, mount_point: str):
		for model in cls.ALL:
			if mount_point in model.mount_points:
				return model

		return None

class PartitionModel(_PartitionModel):
	# system/core/fastboot/fastboot.cpp

	BOOT = _PartitionModel("boot", BOOTLOADER)
	DTBO = _PartitionModel("dtbo", BOOTLOADER)
	RECOVERY = _PartitionModel("recovery", BOOTLOADER)
	MISC = _PartitionModel("misc", BOOTLOADER)
	VBMETA = _PartitionModel("vbmeta", BOOTLOADER)
	VBMETA_SYSTEM = _PartitionModel("vbmeta_system", BOOTLOADER)
	VBMETA_VENDOR = _PartitionModel("vbmeta_vendor", BOOTLOADER)
	VENDOR_BOOT = _PartitionModel("vendor_boot", BOOTLOADER)
	VENDOR_KERNEL_BOOT = _PartitionModel("vendor_kernel_boot", BOOTLOADER)
	INIT_BOOT = _PartitionModel("init_boot", BOOTLOADER)

	SYSTEM = _PartitionModel("system", SSI, ["/system", "/"], Path(""))
	PRODUCT = _PartitionModel("product", SSI)
	SYSTEM_EXT = _PartitionModel("system_ext", SSI)
	SYSTEM_DLKM = _PartitionModel("system_dlkm", SSI)

	VENDOR = _PartitionModel("vendor", TREBLE)
	ODM = _PartitionModel("odm", TREBLE)
	ODM_DLKM = _PartitionModel("odm_dlkm", TREBLE)
	VENDOR_DLKM = _PartitionModel("vendor_dlkm", TREBLE)

	USERDATA = _PartitionModel("data", DATA)
	CACHE = _PartitionModel("cache", DATA)
	METADATA = _PartitionModel("metadata", DATA)
